#
# This file is part of guanaco-ctf.
# Copyright 2021 Diamond Light Source
# Copyright 2021 Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# guanaco-ctf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# guanaco-ctf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with guanaco-ctf. If not, see <http:// www.gnu.org/licenses/>.
#
import mrcfile
import numpy
import guanaco
from matplotlib import pylab
from math import pi


def plot_ctf(
    filename=None,
    q_max=0.5,
    defocus=0,
    spherical_aberration=2.7,
    astigmatism=0,
    astigmatism_angle=0,
    energy=300,
    defocus_spread=None,
    source_spread=0.02,
    chromatic_aberration=2.7,
    energy_spread=0.33e-6,
    current_spread=0.33e-6,
    acceleration_voltage_spread=0.8e-6,
    phase_shift=0,
    component="imag",
    envelope=None,
    ax=None,
    label=None,
    Et_label=None,
    Es_label=None,
    show=False,
):
    """
    Plot the CTF

    Args:
        filename (str): The output filename
        q_max (float): The maximum spatial frequency  (1/A)
        defocus (float): The defocus (df in A)
        spherical_aberration (float): The spherical aberration (Cs in mm)
        astigmatism (float): The astigmatism (A)
        astigmatism_angle (float): The astigmatism angle (deg)
        energy (float): The electron energy (keV)
        defocus_spread (float): The defocus spread (A)
        source_spread (float): The source spread (mrad)
        Cc (float): The chromatic aberrationa (mm)
        dEE (float): dE/E, the fluctuation in the electron energy
        dII (float): dI/I, the fluctuation in the lens current
        dVV (float): dV/V, the fluctuation in the acceleration voltage
        phase_shift (float): The phase shift (rad) - 0 without phase plate, pi/2 with phase plate
        component (str): The bit to plot (real, imag, abs)
        envelope (bool): Plot the envelope function
        ax (object): The axis object (if none create figure)
        label (str): The label for the plot
        Et_label (str): The label for the plot
        Es_label (str): The label for the plot

    """
    # Convert some quantities
    source_spread = source_spread / 1e3  # mrad -> rad
    spherical_aberration = spherical_aberration * 1e7  # mm -> A
    chromatic_aberration = chromatic_aberration * 1e7  # mm -> A
    astigmatism_angle = astigmatism_angle * pi / 180  # deg -> rad
    energy = energy * 1e3  # keV -> eV

    # Compute the defocus spread
    if defocus_spread is None:
        defocus_spread = guanaco.detail.get_defocus_spread(
            Cc=chromatic_aberration,
            dEE=energy_spread,
            dII=current_spread,
            dVV=acceleration_voltage_spread,
        )

    # Check the defocus
    if defocus is None:
        defocus = 0

    # Compute the wavelength
    wavelength = guanaco.detail.get_electron_wavelength(energy)

    # Setup the CTF calculation
    ctf_calculator = guanaco.detail.CTF(
        l=wavelength,
        df=defocus,
        Cs=spherical_aberration,
        Ca=astigmatism,
        Pa=astigmatism_angle,
        dd=defocus_spread,
        theta_c=source_spread,
        phi=phase_shift,
    )

    # Generate the spatial frequencies (1/A)
    q = numpy.arange(0, q_max, q_max / 1000.0)
    theta = numpy.zeros(q.size)

    # Evaluate the ctf
    ctf = ctf_calculator.get_ctf(q, theta)

    # Create the figure
    if ax is None:
        fig, ax = pylab.subplots(figsize=(12, 8))
    else:
        fig = None
    ax.set_xlabel("Spatial frequency (1/A)")
    ax.set_ylabel("CTF")

    # Plot the CTF
    if component is None:
        pass
    elif component == "real":
        if label is None:
            label = "CTF (real)"
        ax.plot(q, numpy.real(ctf), label=label)
    elif component == "imag":
        if label is None:
            label = "CTF (imag)"
        ax.plot(q, numpy.imag(ctf), label=label)
    elif component == "abs":
        if label is None:
            label = "CTF (abs)"
        ax.plot(q, numpy.abs(ctf), label=label)
    else:
        raise ValueError(
            "Expected component in ['real', 'imag', 'abs'], got %s" % component
        )

    # Plot the envelope
    if envelope is not None:

        # Compute the spatial incoherence envelope
        if envelope in ["spatial", "all"]:
            if Es_label is None:
                Es_label = "Es"
            Es = ctf_calculator.get_Es(q, theta)
            ax.plot(q, Es, label=Es_label)

        # Compute the temporal incoherence envelope
        if envelope in ["temporal", "all"]:
            if Et_label is None:
                Et_label = "Et"
            Et = ctf_calculator.get_Et(q)
            ax.plot(q, Et, label=Es_label)

        # Plot the envelope
        ax.legend()

    # Save the figure
    if fig is not None:
        if filename is not None:
            fig.savefig(filename, dpi=300, bbox_inches="tight")
        elif show:
            pylab.show()
        pylab.close(fig)


def generate_ctf(
    filename=None,
    image_size=None,
    pixel_size=1,
    defocus=0,
    num_defocus=None,
    step_defocus=None,
    spherical_aberration=2.7,
    astigmatism=0,
    astigmatism_angle=0,
    energy=300,
    defocus_spread=None,
    source_spread=0.02,
    chromatic_aberration=2.7,
    energy_spread=0.33e-6,
    current_spread=0.33e-6,
    acceleration_voltage_spread=0.8e-6,
    phase_shift=0,
    component="imag",
    envelope=True,
    recentre=False,
):
    """
    Generate a CTF image

    Args:
        filename (str): The output filename
        image_size (tuple): The image size in pixels
        pixel_size (float): The pixel size (A)
        defocus (float): The defocus (df in A)
        spherical_aberration (float): The spherical aberration (Cs in mm)
        astigmatism (float): The astigmatism (A)
        astigmatism_angle (float): The astigmatism angle (deg)
        energy (float): The electron energy (keV)
        defocus_spread (float): The defocus spread (A)
        source_spread (float): The source spread (mrad)
        Cc (float): The chromatic aberrationa (mm)
        dEE (float): dE/E, the fluctuation in the electron energy
        dII (float): dI/I, the fluctuation in the lens current
        dVV (float): dV/V, the fluctuation in the acceleration voltage
        phase_shift (float): The phase shift (rad) - 0 without phase plate, pi/2 with phase plate
        component (str): The bit to plot (real, imag, abs)
        envelope (bool): Plot the envelope function

    """
    # Convert some quantities
    source_spread = source_spread / 1e3  # mrad -> rad
    spherical_aberration = spherical_aberration * 1e7  # mm -> A
    chromatic_aberration = chromatic_aberration * 1e7  # mm -> A
    astigmatism_angle = astigmatism_angle * pi / 180  # deg -> rad
    energy = energy * 1e3  # keV -> eV

    # Compute the defocus spread
    if defocus_spread is None:
        defocus_spread = guanaco.detail.get_defocus_spread(
            Cc=chromatic_aberration,
            dEE=energy_spread,
            dII=current_spread,
            dVV=acceleration_voltage_spread,
        )

    # Check the defocus
    if defocus is None:
        defocus = 0
    if num_defocus is None or num_defocus == 0:
        num_defocus = 1
    if step_defocus is None:
        step_defocus = 0

    # Compute the wavelength
    wavelength = guanaco.detail.get_electron_wavelength(energy)

    # Init output ctf file
    handle = mrcfile.new_mmap(
        filename,
        shape=(num_defocus, image_size[0], image_size[1]),
        mrc_mode=mrcfile.utils.mode_from_dtype(numpy.dtype(numpy.complex64)),
        overwrite=True,
    )

    # Set the voxel size
    handle.voxel_size = pixel_size

    # Loop through the defoci
    for i in range(num_defocus):

        # Set the defocus
        df = defocus + (i - 0.5 * (num_defocus - 1)) * step_defocus
        print("Computing CTF for defocus = %d A" % df)

        # Setup the CTF calculation
        ctf_calculator = guanaco.detail.CTF(
            l=wavelength,
            df=df,
            Cs=spherical_aberration,
            Ca=astigmatism,
            Pa=astigmatism_angle,
            dd=defocus_spread,
            theta_c=source_spread,
            phi=phase_shift,
        )

        # Evaluate the ctf
        ctf = ctf_calculator.get_ctf(image_size[1], image_size[0], pixel_size)

        # Recentre
        if recentre:
            ctf = numpy.fft.fftshift(ctf)

        # Set the CTF
        handle.data[i, :, :] = ctf



# -*- coding: utf-8 -*-
import multiprocessing
import threading

from pybind11_tests import gil_scoped as m


def _run_in_process(target, *args, **kwargs):
    """Runs target in process and returns its exitcode after 10s (None if still alive)."""
    process = multiprocessing.Process(target=target, args=args, kwargs=kwargs)
    process.daemon = True
    try:
        process.start()
        # Do not need to wait much, 10s should be more than enough.
        process.join(timeout=10)
        return process.exitcode
    finally:
        if process.is_alive():
            process.terminate()


def _python_to_cpp_to_python():
    """Calls different C++ functions that come back to Python."""

    class ExtendedVirtClass(m.VirtClass):
        def virtual_func(self):
            pass

        def pure_virtual_func(self):
            pass

    extended = ExtendedVirtClass()
    m.test_callback_py_obj(lambda: None)
    m.test_callback_std_func(lambda: None)
    m.test_callback_virtual_func(extended)
    m.test_callback_pure_virtual_func(extended)


def _python_to_cpp_to_python_from_threads(num_threads, parallel=False):
    """Calls different C++ functions that come back to Python, from Python threads."""
    threads = []
    for _ in range(num_threads):
        thread = threading.Thread(target=_python_to_cpp_to_python)
        thread.daemon = True
        thread.start()
        if parallel:
            threads.append(thread)
        else:
            thread.join()
    for thread in threads:
        thread.join()


# TODO: FIXME, sometimes returns -11 (segfault) instead of 0 on macOS Python 3.9
def test_python_to_cpp_to_python_from_thread():
    """Makes sure there is no GIL deadlock when running in a thread.

    It runs in a separate process to be able to stop and assert if it deadlocks.
    """
    assert _run_in_process(_python_to_cpp_to_python_from_threads, 1) == 0


# TODO: FIXME on macOS Python 3.9
def test_python_to_cpp_to_python_from_thread_multiple_parallel():
    """Makes sure there is no GIL deadlock when running in a thread multiple times in parallel.

    It runs in a separate process to be able to stop and assert if it deadlocks.
    """
    assert _run_in_process(_python_to_cpp_to_python_from_threads, 8, parallel=True) == 0


# TODO: FIXME on macOS Python 3.9
def test_python_to_cpp_to_python_from_thread_multiple_sequential():
    """Makes sure there is no GIL deadlock when running in a thread multiple times sequentially.

    It runs in a separate process to be able to stop and assert if it deadlocks.
    """
    assert (
        _run_in_process(_python_to_cpp_to_python_from_threads, 8, parallel=False) == 0
    )


# TODO: FIXME on macOS Python 3.9
def test_python_to_cpp_to_python_from_process():
    """Makes sure there is no GIL deadlock when using processes.

    This test is for completion, but it was never an issue.
    """
    assert _run_in_process(_python_to_cpp_to_python) == 0


def test_cross_module_gil():
    """Makes sure that the GIL can be acquired by another module from a GIL-released state."""
    m.test_cross_module_gil()  # Should not raise a SIGSEGV



#
# This file is part of guanaco-ctf.
# Copyright 2021 Diamond Light Source
# Copyright 2021 Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# guanaco-ctf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# guanaco-ctf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with guanaco-ctf. If not, see <http:// www.gnu.org/licenses/>.
#

import mrcfile
import numpy
import tempfile
import time
import guanaco
from math import pi


__all__ = ["correct_file", "correct_projections"]


def get_centre(shape, centre, sinogram_order=True):
    if not sinogram_order:
        shape = list(shape)
        shape[0], shape[-2] = shape[-2], shape[0]
    if centre is None:
        centre = numpy.ones(shape[0], dtype="float32") * (shape[-1] / 2.0)
    elif numpy.array(centre).size == 1:
        centre = numpy.ones(shape[0], dtype="float32") * centre
    return centre.astype(dtype="float32", copy=False)


def get_defocus_steps(
    defocus,
    num_defocus=None,
    step_defocus=None,
    width=None,
    centre=None,
    pixel_size=None,
):
    """
    Compute the defocus steps

    Params:
        defocus (float): The defocus at the centre (A)
        num_defocus (int): The number of defocus values
        step_defocus (float): The defocus step (A)
        width (int): The number of pixels in x
        centre (float): The centre of rotation
        pixel_size (float): The pixel size (A)

    Returns:
        The defocus range (inclusive) - min, max, step, num

    """

    # Check the number of defoci
    if num_defocus is None or num_defocus <= 0:
        num_defocus = 1

    # Get the min and max defoci and the step
    if num_defocus == 1:
        min_defocus = defocus
        max_defocus = defocus
        step_defocus = 0
    elif num_defocus > 1:

        # If the step is set then compute min and max directly, otherwise
        # calculate based on the image shape and centre of rotation
        if step_defocus is not None and step_defocus > 0:
            min_defocus = defocus - 0.5 * (num_defocus - 1) * step_defocus
            max_defocus = defocus + 0.5 * (num_defocus - 1) * step_defocus
        else:
            z0 = defocus - centre * pixel_size
            z1 = defocus + centre * pixel_size
            z2 = defocus - (width - centre) * pixel_size
            z3 = defocus + (width - centre) * pixel_size
            min_defocus = min([z0.min(), z1.min(), z2.min(), z3.min()])
            max_defocus = max([z0.max(), z1.max(), z2.max(), z3.max()])
            step_defocus = (max_defocus - min_defocus) / (num_defocus - 1)

    # Return the min and max defocus and defocus step
    return min_defocus, max_defocus, step_defocus, num_defocus


def get_corrected_array(shape, filename_or_array=None):
    """
    Get the corrected array

    """

    # Create a memory mapped file of the correct dimension
    if filename_or_array is None:
        array = numpy.memmap(
            tempfile.TemporaryFile(), mode="w+", dtype="float32", shape=shape
        )
    elif isinstance(filename_or_array, str):
        array = numpy.memmap(filename_or_array, mode="w+", dtype="float32", shape=shape)
    else:
        array = filename_or_array.reshape(shape)

    # Return the array
    return array


def get_ctf(
    width,
    height,
    pixel_size,
    energy,
    defocus,
    spherical_aberration=0,
    astigmatism=0,
    astigmatism_angle=0,
    phase_shift=0,
):
    """
    Get the CTF

    Params:
        width (int): The image width
        height (int): The image height
        shape (tuple): The image size
        pixel_size (float): The pixel size
        energy (float): The electron energy (keV)
        defocus (float): The defocus (A)
        spherical_aberration (float): The spherical aberration (mm)
        astigmatism (float): The 2-fold astigmatism (A)
        astigmatism_angle (float): The angle for 2-fold astigmatism (deg)
        phase_shift (float): The phase shift (deg)

    Returns:
        array: The CTF image

    """
    ctf_calculator = guanaco.detail.CTF(
        l=guanaco.detail.get_electron_wavelength(energy * 1000),
        df=defocus,
        Cs=spherical_aberration * 1e7,
        Ca=astigmatism,
        Pa=astigmatism_angle * pi / 180,
        phi=phase_shift * pi / 180,
    )
    return ctf_calculator.get_ctf_simple(width, height, pixel_size)


def get_ctf_array(
    width,
    height,
    pixel_size,
    energy,
    min_defocus,
    num_defocus,
    step_defocus,
    spherical_aberration=0,
    astigmatism=0,
    astigmatism_angle=0,
    phase_shift=0,
):
    """
    Get the CTF array

    Params:
        width (int): The image width
        height (int): The image height
        pixel_size (float): The pixel size
        energy (float): The electron energy (keV)
        min_defocus (float): The minimum defocus (A)
        num_defocus (int): The number of defoci
        step_defocus (float): The defocus step (A)
        spherical_aberration (float): The spherical aberration (mm)
        astigmatism (float): The 2-fold astigmatism (A)
        astigmatism_angle (float): The angle for 2-fold astigmatism (deg)
        phase_shift (float): The phase shift (deg)

    Returns:
        array: The array of CTF images

    """

    # Precompute the CTF for each defoci
    ctf_array = numpy.zeros((num_defocus, height, width), dtype="complex64")

    # Loop through defoci
    for d in range(num_defocus):

        # Get the defocus
        df = min_defocus + d * step_defocus

        # Generate the ctf
        print("Computing CTF for defocus = %.2f" % df)
        ctf_array[d, :, :] = get_ctf(
            width,
            height,
            pixel_size,
            energy,
            df,
            spherical_aberration,
            astigmatism,
            astigmatism_angle,
            phase_shift,
        )

    # Return the CTF array
    return ctf_array


def correct_projections(
    projections,
    centre=None,
    pixel_size=1,
    energy=None,
    defocus=None,
    num_defocus=None,
    step_defocus=None,
    spherical_aberration=0,
    astigmatism=0,
    astigmatism_angle=0,
    phase_shift=0,
    corrected_output=None,
    device="cpu",
):
    """
    Correct the projections

    Params:
        projections (array): The array of projections
        centre (float): The rotation centre
        pixel_size (float): The pixel size
        energy (float): The electron energy (keV)
        defocus (float): The centre defocus (A)
        num_defocus (int): The number of defoci
        step_defocus (float): The defocus step (A)
        spherical_aberration (float): The spherical aberration (mm)
        astigmatism (float): The 2-fold astigmatism (A)
        astigmatism_angle (float): The angle for 2-fold astigmatism (deg)
        phase_shift (float): The phase shift (deg)
        corrected_output (object): The output data
        device (str): The cpu or gpu

    Returns:
        projections, min_defocus, max_defocus

    """

    # If we have no defocus set then do nothing. Otherwise correct them
    if defocus is not None:

        # Compute the defocus steps
        min_defocus, max_defocus, step_defocus, num_defocus = get_defocus_steps(
            defocus, num_defocus, step_defocus, projections.shape[2], centre, pixel_size
        )

        # Check consistency
        assert (
            abs(min_defocus + (num_defocus - 1) * step_defocus - max_defocus) < 1e-7
        ), "Inconsistent defocus steps"

        # The shape of the corrected projection array
        corrected_shape = (
            projections.shape[0],
            num_defocus,
            projections.shape[1],
            projections.shape[2],
        )

        # Init the corrected projections array
        corrected_projections = get_corrected_array(corrected_shape, corrected_output)

        # Precompute the CTF for each defoci
        ctf_array = get_ctf_array(
            projections.shape[2],
            projections.shape[1],
            pixel_size,
            energy,
            min_defocus,
            num_defocus,
            step_defocus,
            spherical_aberration,
            astigmatism,
            astigmatism_angle,
            phase_shift,
        )

        # Loop through all the projections and defoci and perform the CTF
        # correction
        for z in range(projections.shape[0]):
            image = projections[z, :, :]
            print(
                "Correcting image %d/%d with %d defoci"
                % (z + 1, projections.shape[0], len(ctf_array))
            )
            guanaco.detail.corr(
                image, ctf_array, corrected_projections[z, :, :, :], device
            )

        # Remove the dimension for corrections if only one correction
        if corrected_projections.shape[1] == 1:
            corrected_projections = corrected_projections.reshape(projections.shape)

        # Set the projections to be the corrected projections
        projections = corrected_projections

    else:
        defocus = 0
        min_defocus = 0
        max_defocus = 0

    # Return the projections with the min and max relative defoci
    return (projections, min_defocus - defocus, max_defocus - defocus)


def correct_file(
    input_filename,
    output_filename,
    centre=None,
    energy=None,
    defocus=None,
    num_defocus=None,
    step_defocus=None,
    spherical_aberration=None,
    astigmatism=None,
    astigmatism_angle=None,
    phase_shift=None,
    device="cpu",
):
    """
    Perform the CTF correction

    Params:
        input_filename (str): The input projections filename
        output_filename (str): The output corrected projections filename
        centre (float): The rotation centre
        energy (float): The electron energy (keV)
        defocus (float): The centre defocus (A)
        num_defocus (int): The number of defoci
        step_defocus (float): The defocus step (A)
        spherical_aberration (float): The spherical aberration (mm)
        astigmatism (float): The 2-fold astigmatism (A)
        astigmatism_angle (float): The angle for 2-fold astigmatism (deg)
        phase_shift (float): The phase shift (deg)
        device (str): The cpu or gpu

    """

    start_time = time.time()

    def read_projection_metadata(infile):

        # Read the voxel size
        voxel_size = infile.voxel_size
        print("Voxel size: ", infile.voxel_size)

        # Read the angles
        assert (
            infile.data.shape[0] == infile.extended_header.shape[0]
        ), "Inconsistent data/header array shape"
        angles = numpy.zeros(infile.extended_header.shape[0], dtype=numpy.float32)
        for i in range(infile.extended_header.shape[0]):
            angles[i] = numpy.deg2rad(infile.extended_header[i]["Alpha tilt"])
            print("Image %d; angle %.4f deg" % (i + 1, numpy.rad2deg(angles[i])))

        # Return metadata
        return angles, voxel_size

    def open_corrected_file(output_filename, shape, voxel_size):

        try:
            header_dtype = mrcfile.dtypes.FEI1_EXTENDED_HEADER_DTYPE
        except Exception:
            header_dtype = mrcfile.dtypes.get_ext_header_dtype(b"FEI1")

        # Set the data type
        dtype = numpy.dtype(numpy.float32)

        # Create the extended header
        extended_header = numpy.zeros(shape=shape[0:2], dtype=header_dtype)

        # Open the file
        outfile = mrcfile.new_mmap(
            output_filename,
            shape=(0, 0, 0, 0),
            mrc_mode=mrcfile.utils.mode_from_dtype(dtype),
            overwrite=True,
        )

        # Some hacks to set the extended header without having to resize whole file.
        outfile._check_writeable()
        outfile._close_data()
        outfile._extended_header = extended_header
        outfile.header.nsymbt = extended_header.nbytes
        outfile.header.exttyp = "FEI1"
        outfile._open_memmap(dtype, shape)
        outfile.update_header_from_data()
        outfile.flush()

        # Set the voxel size
        outfile.voxel_size = voxel_size

        # Return the handle
        return outfile

    # Open the input file
    print("Reading %s" % input_filename)
    with mrcfile.mmap(input_filename) as infile:

        # Get the projection metadata
        angles, voxel_size = read_projection_metadata(infile)

        # Get the pixel size
        assert voxel_size["x"] == voxel_size["y"]
        pixel_size = voxel_size["x"]

        # Get the projection data
        projections = infile.data

        # Get the rotation centre
        centre = get_centre(projections.shape, centre, False)

        # Set the number of defoci
        if num_defocus is None or num_defocus == 0:
            num_defocus = 1
            step_defocus = 0

        # The shape of the corrected projection array
        output_shape = (
            projections.shape[0],
            num_defocus,
            projections.shape[1],
            projections.shape[2],
        )

        # Open the output file
        print("Writing corrected projections to %s" % output_filename)
        with open_corrected_file(output_filename, output_shape, voxel_size) as outfile:

            # Get the corrected data
            corrected = outfile.data

            # Reconstruct
            _, min_defocus, max_defocus = correct_projections(
                projections,
                centre,
                pixel_size,
                energy,
                defocus,
                num_defocus,
                step_defocus,
                spherical_aberration,
                astigmatism,
                astigmatism_angle,
                phase_shift,
                corrected,
                device,
            )

            # Compute step defocus if necessary
            if step_defocus is None:
                if num_defocus == 1:
                    step_defocus = 0
                else:
                    step_defocus = (max_defocus - min_defocus) / (num_defocus - 1)

            # Set the corrected data metadata
            for j in range(output_shape[0]):
                for i in range(output_shape[1]):
                    outfile.extended_header[j, i]["Alpha tilt"] = numpy.rad2deg(
                        angles[j]
                    )
                    outfile.extended_header[j, i]["Defocus"] = (
                        min_defocus + i * step_defocus
                    )

    print("Time: %.2f seconds" % (time.time() - start_time))



#
# This file is part of guanaco-ctf.
# Copyright 2021 Diamond Light Source
# Copyright 2021 Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# guanaco-ctf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# guanaco-ctf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with guanaco-ctf. If not, see <http:// www.gnu.org/licenses/>.
#
from guanaco_ext import *  # noqa



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re

import ghapi.core


ENTRY = re.compile(
    r"""
    Suggested \s changelog \s entry:
    .*
    ```rst
    \s*
    (.*?)
    \s*
    ```
""",
    re.DOTALL | re.VERBOSE,
)


api = ghapi.core.GhApi(owner="pybind", repo="pybind11")

issues = api.issues.list_for_repo(labels="needs changelog", state="closed")
missing = []

for issue in issues:
    changelog = ENTRY.findall(issue.body)
    if changelog:
        (msg,) = changelog
        if not msg.startswith("* "):
            msg = "* " + msg
        if not msg.endswith("."):
            msg += "."

        print(msg)
        print(f"  `#{issue.number} <{issue.html_url}>`_\n")

    else:
        missing.append(issue)

if missing:
    print()
    print("-" * 30)
    print()

    for issue in missing:
        print(f"Missing: {issue.title}")
        print(f"  {issue.html_url}")



from __future__ import absolute_import, division, print_function, unicode_literals

import guanaco.detail

import numpy
import multiprocessing as mp
import concurrent.futures as cf

# FIXME DOCSTRING
def get_ncore_slices(axis_size, ncore=None, nchunk=None):
    # default ncore to max (also defaults ncore == 0)
    if not ncore:
        ncore = mp.cpu_count()
    if nchunk is None:
        # calculate number of slices to send to each GPU
        chunk_size = axis_size // ncore
        leftover = axis_size % ncore
        sizes = numpy.ones(ncore, dtype="int") * chunk_size
        # evenly distribute leftover across workers
        sizes[:leftover] += 1
        offsets = numpy.zeros(ncore + 1, dtype="int")
        offsets[1:] = numpy.cumsum(sizes)
        slices = [
            numpy.s_[offsets[i] : offsets[i + 1]] for i in range(offsets.shape[0] - 1)
        ]
    elif nchunk == 0:
        # nchunk == 0 is a special case, we will collapse the dimension
        slices = [numpy.s_[i] for i in range(axis_size)]
    else:
        # calculate offsets based on chunk size
        slices = [
            numpy.s_[offset : offset + nchunk] for offset in range(0, axis_size, nchunk)
        ]
    return ncore, slices


def reconstruction_dispatcher(
    sinogram,
    reconstruction,
    centre,
    angles,
    weights=None,
    pixel_size=1,
    min_defocus=0,
    max_defocus=0,
    transform=None,
    device="cpu",
    ncore=None,
    nchunk=None,
    gpu_list=None,
):

    # Unpack arguments
    nslices = sinogram.shape[0]

    # Set the number of cores in the case of GPU
    if device == "gpu" and ncore is None:
        ncore = 1

    # Compute the number of cores
    if device == "gpu" and gpu_list is not None:
        ngpu = len(gpu_list)
        ncore, slices = get_ncore_slices(nslices, ngpu, nchunk)
        assert ncore == len(slices), "# cores must equal # slices"
        assert ncore == ngpu, "# cores must equal # gpus"
    else:
        ncore, slices = get_ncore_slices(nslices, ncore, nchunk)
        assert ncore == len(slices), "# cores must equal # slices"
        gpu_list = [None] * ncore

    # If only one core then run on this thread, otherwise spawn other threads
    if ncore == 1:
        for s in slices:
            reconstruction_worker(
                sinogram[s],
                reconstruction[s],
                centre[s],
                angles,
                weights,
                pixel_size,
                min_defocus,
                max_defocus,
                transform,
                device,
                None,
            )
    else:
        with cf.ThreadPoolExecutor(ncore) as e:
            for gpu, s in zip(gpu_list, slices):
                e.submit(
                    reconstruction_worker,
                    sinogram[s],
                    reconstruction[s],
                    centre[s],
                    angles,
                    weights,
                    pixel_size,
                    min_defocus,
                    max_defocus,
                    transform,
                    device,
                    gpu,
                )


def reconstruction_worker(
    sinogram,
    reconstruction,
    centre,
    angles,
    weights,
    pixel_size,
    min_defocus,
    max_defocus,
    transform,
    device,
    gpu_index,
):

    if gpu_index is None:
        gpu_index = 0

    nslices = sinogram.shape[0]

    for i in range(nslices):
        sino = sinogram[i]

        # Transform the corrected projections
        if transform == "minus":
            sino = -sino

        # Get as contiguous array
        sino = numpy.ascontiguousarray(sino)

        # Set weights
        if weights is not None:
            if len(sino.shape) == 3:
                for j in range(sino.shape[1]):
                    sino[:, j, :] *= weights[j]
            else:
                for j in range(sino.shape[1]):
                    sino[j, :] *= weights[j]

        print("Reconstructing slice %d/%d" % (i + 1, nslices))
        guanaco.detail.recon(
            sino,
            reconstruction[i],
            angles,
            centre[i],
            pixel_size,
            min_defocus,
            max_defocus,
            device,
            gpu_index,
        )



# -*- coding: utf-8 -*-
import pytest

m = pytest.importorskip("pybind11_tests.smart_ptr")
from pybind11_tests import ConstructorStats  # noqa: E402


def test_smart_ptr(capture):
    # Object1
    for i, o in enumerate(
        [m.make_object_1(), m.make_object_2(), m.MyObject1(3)], start=1
    ):
        assert o.getRefCount() == 1
        with capture:
            m.print_object_1(o)
            m.print_object_2(o)
            m.print_object_3(o)
            m.print_object_4(o)
        assert capture == "MyObject1[{i}]\n".format(i=i) * 4

    for i, o in enumerate(
        [m.make_myobject1_1(), m.make_myobject1_2(), m.MyObject1(6), 7], start=4
    ):
        print(o)
        with capture:
            if not isinstance(o, int):
                m.print_object_1(o)
                m.print_object_2(o)
                m.print_object_3(o)
                m.print_object_4(o)
            m.print_myobject1_1(o)
            m.print_myobject1_2(o)
            m.print_myobject1_3(o)
            m.print_myobject1_4(o)

        times = 4 if isinstance(o, int) else 8
        assert capture == "MyObject1[{i}]\n".format(i=i) * times

    cstats = ConstructorStats.get(m.MyObject1)
    assert cstats.alive() == 0
    expected_values = ["MyObject1[{}]".format(i) for i in range(1, 7)] + [
        "MyObject1[7]"
    ] * 4
    assert cstats.values() == expected_values
    assert cstats.default_constructions == 0
    assert cstats.copy_constructions == 0
    # assert cstats.move_constructions >= 0 # Doesn't invoke any
    assert cstats.copy_assignments == 0
    assert cstats.move_assignments == 0

    # Object2
    for i, o in zip(
        [8, 6, 7], [m.MyObject2(8), m.make_myobject2_1(), m.make_myobject2_2()]
    ):
        print(o)
        with capture:
            m.print_myobject2_1(o)
            m.print_myobject2_2(o)
            m.print_myobject2_3(o)
            m.print_myobject2_4(o)
        assert capture == "MyObject2[{i}]\n".format(i=i) * 4

    cstats = ConstructorStats.get(m.MyObject2)
    assert cstats.alive() == 1
    o = None
    assert cstats.alive() == 0
    assert cstats.values() == ["MyObject2[8]", "MyObject2[6]", "MyObject2[7]"]
    assert cstats.default_constructions == 0
    assert cstats.copy_constructions == 0
    # assert cstats.move_constructions >= 0 # Doesn't invoke any
    assert cstats.copy_assignments == 0
    assert cstats.move_assignments == 0

    # Object3
    for i, o in zip(
        [9, 8, 9], [m.MyObject3(9), m.make_myobject3_1(), m.make_myobject3_2()]
    ):
        print(o)
        with capture:
            m.print_myobject3_1(o)
            m.print_myobject3_2(o)
            m.print_myobject3_3(o)
            m.print_myobject3_4(o)
        assert capture == "MyObject3[{i}]\n".format(i=i) * 4

    cstats = ConstructorStats.get(m.MyObject3)
    assert cstats.alive() == 1
    o = None
    assert cstats.alive() == 0
    assert cstats.values() == ["MyObject3[9]", "MyObject3[8]", "MyObject3[9]"]
    assert cstats.default_constructions == 0
    assert cstats.copy_constructions == 0
    # assert cstats.move_constructions >= 0 # Doesn't invoke any
    assert cstats.copy_assignments == 0
    assert cstats.move_assignments == 0

    # Object
    cstats = ConstructorStats.get(m.Object)
    assert cstats.alive() == 0
    assert cstats.values() == []
    assert cstats.default_constructions == 10
    assert cstats.copy_constructions == 0
    # assert cstats.move_constructions >= 0 # Doesn't invoke any
    assert cstats.copy_assignments == 0
    assert cstats.move_assignments == 0

    # ref<>
    cstats = m.cstats_ref()
    assert cstats.alive() == 0
    assert cstats.values() == ["from pointer"] * 10
    assert cstats.default_constructions == 30
    assert cstats.copy_constructions == 12
    # assert cstats.move_constructions >= 0 # Doesn't invoke any
    assert cstats.copy_assignments == 30
    assert cstats.move_assignments == 0


def test_smart_ptr_refcounting():
    assert m.test_object1_refcounting()


def test_unique_nodelete():
    o = m.MyObject4(23)
    assert o.value == 23
    cstats = ConstructorStats.get(m.MyObject4)
    assert cstats.alive() == 1
    del o
    assert cstats.alive() == 1
    m.MyObject4.cleanup_all_instances()
    assert cstats.alive() == 0


def test_unique_nodelete4a():
    o = m.MyObject4a(23)
    assert o.value == 23
    cstats = ConstructorStats.get(m.MyObject4a)
    assert cstats.alive() == 1
    del o
    assert cstats.alive() == 1
    m.MyObject4a.cleanup_all_instances()
    assert cstats.alive() == 0


def test_unique_deleter():
    m.MyObject4a(0)
    o = m.MyObject4b(23)
    assert o.value == 23
    cstats4a = ConstructorStats.get(m.MyObject4a)
    assert cstats4a.alive() == 2
    cstats4b = ConstructorStats.get(m.MyObject4b)
    assert cstats4b.alive() == 1
    del o
    assert cstats4a.alive() == 1  # Should now only be one leftover
    assert cstats4b.alive() == 0  # Should be deleted
    m.MyObject4a.cleanup_all_instances()
    assert cstats4a.alive() == 0
    assert cstats4b.alive() == 0


def test_large_holder():
    o = m.MyObject5(5)
    assert o.value == 5
    cstats = ConstructorStats.get(m.MyObject5)
    assert cstats.alive() == 1
    del o
    assert cstats.alive() == 0


def test_shared_ptr_and_references():
    s = m.SharedPtrRef()
    stats = ConstructorStats.get(m.A)
    assert stats.alive() == 2

    ref = s.ref  # init_holder_helper(holder_ptr=false, owned=false)
    assert stats.alive() == 2
    assert s.set_ref(ref)
    with pytest.raises(RuntimeError) as excinfo:
        assert s.set_holder(ref)
    assert "Unable to cast from non-held to held instance" in str(excinfo.value)

    copy = s.copy  # init_holder_helper(holder_ptr=false, owned=true)
    assert stats.alive() == 3
    assert s.set_ref(copy)
    assert s.set_holder(copy)

    holder_ref = s.holder_ref  # init_holder_helper(holder_ptr=true, owned=false)
    assert stats.alive() == 3
    assert s.set_ref(holder_ref)
    assert s.set_holder(holder_ref)

    holder_copy = s.holder_copy  # init_holder_helper(holder_ptr=true, owned=true)
    assert stats.alive() == 3
    assert s.set_ref(holder_copy)
    assert s.set_holder(holder_copy)

    del ref, copy, holder_ref, holder_copy, s
    assert stats.alive() == 0


def test_shared_ptr_from_this_and_references():
    s = m.SharedFromThisRef()
    stats = ConstructorStats.get(m.B)
    assert stats.alive() == 2

    ref = s.ref  # init_holder_helper(holder_ptr=false, owned=false, bad_wp=false)
    assert stats.alive() == 2
    assert s.set_ref(ref)
    assert s.set_holder(
        ref
    )  # std::enable_shared_from_this can create a holder from a reference

    bad_wp = s.bad_wp  # init_holder_helper(holder_ptr=false, owned=false, bad_wp=true)
    assert stats.alive() == 2
    assert s.set_ref(bad_wp)
    with pytest.raises(RuntimeError) as excinfo:
        assert s.set_holder(bad_wp)
    assert "Unable to cast from non-held to held instance" in str(excinfo.value)

    copy = s.copy  # init_holder_helper(holder_ptr=false, owned=true, bad_wp=false)
    assert stats.alive() == 3
    assert s.set_ref(copy)
    assert s.set_holder(copy)

    holder_ref = (
        s.holder_ref
    )  # init_holder_helper(holder_ptr=true, owned=false, bad_wp=false)
    assert stats.alive() == 3
    assert s.set_ref(holder_ref)
    assert s.set_holder(holder_ref)

    holder_copy = (
        s.holder_copy
    )  # init_holder_helper(holder_ptr=true, owned=true, bad_wp=false)
    assert stats.alive() == 3
    assert s.set_ref(holder_copy)
    assert s.set_holder(holder_copy)

    del ref, bad_wp, copy, holder_ref, holder_copy, s
    assert stats.alive() == 0

    z = m.SharedFromThisVirt.get()
    y = m.SharedFromThisVirt.get()
    assert y is z


def test_move_only_holder():
    a = m.TypeWithMoveOnlyHolder.make()
    b = m.TypeWithMoveOnlyHolder.make_as_object()
    stats = ConstructorStats.get(m.TypeWithMoveOnlyHolder)
    assert stats.alive() == 2
    del b
    assert stats.alive() == 1
    del a
    assert stats.alive() == 0


def test_holder_with_addressof_operator():
    # this test must not throw exception from c++
    a = m.TypeForHolderWithAddressOf.make()
    a.print_object_1()
    a.print_object_2()
    a.print_object_3()
    a.print_object_4()

    stats = ConstructorStats.get(m.TypeForHolderWithAddressOf)
    assert stats.alive() == 1

    np = m.TypeForHolderWithAddressOf.make()
    assert stats.alive() == 2
    del a
    assert stats.alive() == 1
    del np
    assert stats.alive() == 0

    b = m.TypeForHolderWithAddressOf.make()
    c = b
    assert b.get() is c.get()
    assert stats.alive() == 1

    del b
    assert stats.alive() == 1

    del c
    assert stats.alive() == 0


def test_move_only_holder_with_addressof_operator():
    a = m.TypeForMoveOnlyHolderWithAddressOf.make()
    a.print_object()

    stats = ConstructorStats.get(m.TypeForMoveOnlyHolderWithAddressOf)
    assert stats.alive() == 1

    a.value = 42
    assert a.value == 42

    del a
    assert stats.alive() == 0


def test_smart_ptr_from_default():
    instance = m.HeldByDefaultHolder()
    with pytest.raises(RuntimeError) as excinfo:
        m.HeldByDefaultHolder.load_shared_ptr(instance)
    assert (
        "Unable to load a custom holder type from a "
        "default-holder instance" in str(excinfo.value)
    )


def test_shared_ptr_gc():
    """#187: issue involving std::shared_ptr<> return value policy & garbage collection"""
    el = m.ElementList()
    for i in range(10):
        el.add(m.ElementA(i))
    pytest.gc_collect()
    for i, v in enumerate(el.get()):
        assert i == v.value()



#
# This file is part of guanaco-ctf.
# Copyright 2021 Diamond Light Source
# Copyright 2021 Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# guanaco-ctf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# guanaco-ctf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with guanaco-ctf. If not, see <http:// www.gnu.org/licenses/>.
#
import argparse
import guanaco
import guanaco.plot


def add_argument(parser, name):
    """
    Add common command line arguments

    """
    {
        "energy": lambda: (
            parser.add_argument(
                "--E",
                dest="energy",
                default=300,
                type=float,
                help="The energy of the incident electrons (keV)",
            )
        ),
        "defocus": lambda: (
            parser.add_argument(
                "--df",
                dest="defocus",
                default=None,
                type=float,
                help="The defocus at the rotation axis (A)",
            )
        ),
        "num_defocus": lambda: (
            parser.add_argument(
                "--ndf",
                dest="num_defocus",
                default=None,
                type=int,
                help="The number of defocus steps to use",
            )
        ),
        "step_defocus": lambda: (
            parser.add_argument(
                "--sdf",
                dest="step_defocus",
                default=None,
                type=int,
                help="The defocus step to use",
            )
        ),
        "spherical_aberration": lambda: (
            parser.add_argument(
                "--Cs",
                dest="spherical_aberration",
                default=0,
                type=float,
                help="The spherical aberration (mm)",
            )
        ),
        "astigmatism": lambda: (
            parser.add_argument(
                "--Ca",
                dest="astigmatism",
                default=0,
                type=float,
                help="The 2-fold astigmatism (A)",
            )
        ),
        "astigmatism_angle": lambda: (
            parser.add_argument(
                "--Pa",
                dest="astigmatism_angle",
                default=0,
                type=float,
                help="The angle for the 2-fold astigmatism (deg)",
            )
        ),
        "phase_shift": lambda: (
            parser.add_argument(
                "--phi",
                dest="phase_shift",
                default=0,
                type=float,
                help="The phase shift to apply to the CTF (deg)",
            )
        ),
        "defocus_spread": lambda: (
            parser.add_argument(
                "--d",
                "--defocus_spread",
                dest="defocus_spread",
                type=float,
                default=None,
                help="The defocus spread (A)",
            )
        ),
        "source_spread": lambda: (
            parser.add_argument(
                "--theta_c",
                "--source_spread",
                dest="source_spread",
                type=float,
                default=0.02,
                help="The source spread (mrad)",
            )
        ),
        "chromatic_aberration": lambda: (
            parser.add_argument(
                "--Cc",
                "--chromatic_aberration",
                dest="chromatic_aberration",
                type=float,
                default=2.7,
                help="The chromatic aberration (mm)",
            )
        ),
        "energy_spread": lambda: (
            parser.add_argument(
                "--dEE",
                "--energy_spread",
                dest="energy_spread",
                type=float,
                default=0.33e-6,
                help="The energy spread dE / E",
            )
        ),
        "acceleration_voltage_spread": lambda: (
            parser.add_argument(
                "--dVV",
                "--acceleration_voltage_spread",
                dest="acceleration_voltage_spread",
                type=float,
                default=0.8e-6,
                help="The acceleration voltage spread dV / V",
            )
        ),
        "current_spread": lambda: (
            parser.add_argument(
                "--dII",
                "--current_spread",
                dest="current_spread",
                type=float,
                default=0.33e-6,
                help="The current spread dI / I",
            )
        ),
    }[name]()


def main(args=None):
    """
    Do the reconstruction

    """

    # Create the command line parser
    parser = argparse.ArgumentParser(description="Reconstruct")

    parser.add_argument(
        "-o",
        dest="output",
        default="rec.mrc",
        help="The output file containing the reconstructed projections.",
    )

    parser.add_argument(
        "-i",
        dest="input",
        default=None,
        required=True,
        help="The input file containing a stack of projection images.",
    )

    parser.add_argument(
        "--corrected-filename",
        dest="corrected_filename",
        default="GUANACO_CORRECTED.dat",
        help="The intermediate file containing the CTF corrected projections",
    )

    parser.add_argument(
        "-d,--device",
        dest="device",
        default="cpu",
        choices=["cpu", "gpu"],
        help="Use either the CPU or GPU",
    )

    parser.add_argument(
        "--transform",
        dest="transform",
        default=None,
        choices=["none", "minus"],
        help="Set the transform to use on the corrected projections",
    )

    parser.add_argument(
        "--angular_weights",
        dest="angular_weights",
        default=False,
        type=bool,
        help="Set whether or not to use angular weights",
    )

    parser.add_argument(
        "-n,--ncore",
        dest="ncore",
        default=None,
        type=int,
        help="Set the number of cores to use",
    )

    parser.add_argument(
        "--chunk_size",
        dest="chunk_size",
        default=None,
        type=int,
        help="Set the number of rows to reconstruct",
    )

    parser.add_argument(
        "--start_angle",
        dest="start_angle",
        default=None,
        type=float,
        help="The starting angle in degrees.",
    )

    parser.add_argument(
        "--step_angle",
        dest="step_angle",
        default=None,
        type=float,
        help="The step angle in degrees.",
    )

    parser.add_argument(
        "--pixel_size",
        dest="pixel_size",
        default=None,
        type=float,
        help="The pixel size in A",
    )

    parser.add_argument(
        "--centre",
        dest="centre",
        default=None,
        type=float,
        help="The rotation centre in pixels.",
    )

    parser.add_argument(
        "--method",
        dest="method",
        default="FBP_CTF",
        choices=["FBP_CTF", "FBP", "SIRT", "SART", "CGLS", "EM"],
        help="""
            Select the reconstruction algorithm. The FBP_CTF algorithm uses 3D
            CTF correction with FBP, other algorithms require tomopy to be
            installed""",
    )

    parser.add_argument(
        "--num_iter",
        dest="num_iter",
        default=50,
        type=int,
        help="The number of iterations to use in SIRT/SART/CGLS/EM",
    )

    # Add some common arguments
    add_argument(parser, "energy")
    add_argument(parser, "defocus")
    add_argument(parser, "num_defocus")
    add_argument(parser, "step_defocus")
    add_argument(parser, "spherical_aberration")
    add_argument(parser, "astigmatism")
    add_argument(parser, "astigmatism_angle")
    add_argument(parser, "phase_shift")

    # Parse the arguments
    args = parser.parse_args(args=args)

    # Do the reconstruction
    guanaco.reconstruct_file(
        input_filename=args.input,
        output_filename=args.output,
        corrected_filename=args.corrected_filename,
        start_angle=args.start_angle,
        step_angle=args.step_angle,
        pixel_size=args.pixel_size,
        centre=args.centre,
        energy=args.energy,
        defocus=args.defocus,
        num_defocus=args.num_defocus,
        step_defocus=args.step_defocus,
        spherical_aberration=args.spherical_aberration,
        astigmatism=args.astigmatism,
        astigmatism_angle=args.astigmatism_angle,
        phase_shift=args.phase_shift,
        device=args.device,
        ncore=args.ncore,
        transform=args.transform,
        angular_weights=args.angular_weights,
        chunk_size=args.chunk_size,
        method=args.method,
        num_iter=args.num_iter,
    )


def plot_ctf(args=None):
    """
    Plot the CTF

    Params:
        args (list): The input command line arguments

    """

    # Create the command line parser
    parser = argparse.ArgumentParser(description="Plot the CTF")

    parser.add_argument(
        "-o",
        dest="output",
        default="ctf.png",
        help="The output image file for the CTF plot",
    )

    parser.add_argument(
        "--q_max",
        dest="q_max",
        type=float,
        default=0.5,
        help="The maximum spatial frequency (1/A)",
    )

    parser.add_argument(
        "--component",
        dest="component",
        type=str,
        choices=["real", "imag", "abs"],
        default="imag",
        help="The component to plot",
    )

    parser.add_argument(
        "--envelope",
        dest="envelope",
        type=str,
        choices=["spatial", "temporal", "all"],
        default=None,
        help="Plot the CTF envelope",
    )

    # Add some common arguments
    add_argument(parser, "energy")
    add_argument(parser, "defocus")
    add_argument(parser, "spherical_aberration")
    add_argument(parser, "astigmatism")
    add_argument(parser, "astigmatism_angle")
    add_argument(parser, "phase_shift")
    add_argument(parser, "defocus_spread")
    add_argument(parser, "source_spread")
    add_argument(parser, "chromatic_aberration")
    add_argument(parser, "energy_spread")
    add_argument(parser, "acceleration_voltage_spread")
    add_argument(parser, "current_spread")

    # Parse the arguments
    args = parser.parse_args(args=args)

    # Do the reconstruction
    guanaco.plot.plot_ctf(
        filename=args.output,
        q_max=args.q_max,
        defocus=args.defocus,
        spherical_aberration=args.spherical_aberration,
        astigmatism=args.astigmatism,
        astigmatism_angle=args.astigmatism_angle,
        energy=args.energy,
        defocus_spread=args.defocus_spread,
        source_spread=args.source_spread,
        chromatic_aberration=args.chromatic_aberration,
        energy_spread=args.energy_spread,
        current_spread=args.current_spread,
        acceleration_voltage_spread=args.acceleration_voltage_spread,
        phase_shift=args.phase_shift,
        component=args.component,
        envelope=args.envelope,
    )


def generate_ctf(args=None):
    """
    Generate the CTF image

    """

    # Create the command line parser
    parser = argparse.ArgumentParser(description="Generate the CTF image")

    parser.add_argument(
        "-o",
        dest="output",
        default="ctf.mrc",
        help="The output image file for the CTF image",
    )

    parser.add_argument(
        "--size",
        "--image_size",
        dest="image_size",
        type=lambda x: list(map(int, x.split(","))),
        default=None,
        required=True,
        help="The image size",
    )

    parser.add_argument(
        "--px",
        "--pixel_size",
        dest="pixel_size",
        type=float,
        default=1,
        help="The pixel size",
    )

    parser.add_argument(
        "--recentre",
        dest="recentre",
        type=bool,
        default=False,
        help="Recentre the CTF in Fourier space",
    )

    # Add some common arguments
    add_argument(parser, "energy")
    add_argument(parser, "defocus")
    add_argument(parser, "num_defocus")
    add_argument(parser, "step_defocus")
    add_argument(parser, "spherical_aberration")
    add_argument(parser, "astigmatism")
    add_argument(parser, "astigmatism_angle")
    add_argument(parser, "phase_shift")
    add_argument(parser, "defocus_spread")
    add_argument(parser, "source_spread")
    add_argument(parser, "chromatic_aberration")
    add_argument(parser, "energy_spread")
    add_argument(parser, "acceleration_voltage_spread")
    add_argument(parser, "current_spread")

    # Parse the arguments
    args = parser.parse_args(args=args)

    # Do the reconstruction
    guanaco.plot.generate_ctf(
        filename=args.output,
        image_size=args.image_size,
        pixel_size=args.pixel_size,
        defocus=args.defocus,
        num_defocus=args.num_defocus,
        step_defocus=args.step_defocus,
        spherical_aberration=args.spherical_aberration,
        astigmatism=args.astigmatism,
        astigmatism_angle=args.astigmatism_angle,
        energy=args.energy,
        defocus_spread=args.defocus_spread,
        source_spread=args.source_spread,
        chromatic_aberration=args.chromatic_aberration,
        energy_spread=args.energy_spread,
        current_spread=args.current_spread,
        acceleration_voltage_spread=args.acceleration_voltage_spread,
        phase_shift=args.phase_shift,
        recentre=args.recentre,
    )


def correct(args=None):
    """
    Do the CTF correction

    """

    # Create the command line parser
    parser = argparse.ArgumentParser(description="Perform CTF correction")

    parser.add_argument(
        "-o",
        dest="output",
        default="corrected.mrc",
        type=str,
        help="The output file containing the corrected projections.",
    )

    parser.add_argument(
        "-i",
        dest="input",
        default=None,
        required=True,
        type=str,
        help="The input file containing a stack of projection images.",
    )

    parser.add_argument(
        "-d,--device",
        dest="device",
        default="cpu",
        choices=["cpu", "gpu"],
        help="Use either the CPU or GPU",
    )

    parser.add_argument(
        "--centre",
        dest="centre",
        default=None,
        type=float,
        help="The rotation centre in pixels.",
    )

    # Add some common arguments
    add_argument(parser, "energy")
    add_argument(parser, "defocus")
    add_argument(parser, "num_defocus")
    add_argument(parser, "step_defocus")
    add_argument(parser, "spherical_aberration")
    add_argument(parser, "astigmatism")
    add_argument(parser, "astigmatism_angle")
    add_argument(parser, "phase_shift")

    # Parse the arguments
    args = parser.parse_args(args=args)

    # Do the reconstruction
    guanaco.correct_file(
        input_filename=args.input,
        output_filename=args.output,
        centre=args.centre,
        energy=args.energy,
        defocus=args.defocus,
        num_defocus=args.num_defocus,
        step_defocus=args.step_defocus,
        spherical_aberration=args.spherical_aberration,
        astigmatism=args.astigmatism,
        astigmatism_angle=args.astigmatism_angle,
        phase_shift=args.phase_shift,
        device=args.device,
    )



#
# This file is part of guanaco-ctf.
# Copyright 2021 Diamond Light Source
# Copyright 2021 Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# guanaco-ctf is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# guanaco-ctf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with guanaco-ctf. If not, see <http:// www.gnu.org/licenses/>.
#


import mrcfile
import numpy
import time
import guanaco.detail
import guanaco.detail.mp

__all__ = ["reconstruct_file", "reconstruct"]


def get_centre(shape, centre, sinogram_order=True):
    """
    Get the centre of the sinogram

    Params:
        shape (tuple): The sinogram shape
        centre (object): The input centre
        sinogram_order (bool): The order of the input data

    Returns:
        array: The centre array

    """
    if not sinogram_order:
        shape = list(shape)
        shape[0], shape[-2] = shape[-2], shape[0]
    if centre is None:
        centre = numpy.ones(shape[0], dtype="float32") * (shape[-1] * 0.5)
    elif numpy.array(centre).size == 1:
        centre = numpy.ones(shape[0], dtype="float32") * centre
    return centre.astype(dtype="float32", copy=False)


def compute_angular_weights(angles):
    """
    Compute angular weights for the projections

    Params:
        angles (float): The projection angles

    Returns:
        array: The weights

    """

    # Sort the angle indices
    index = numpy.argsort(angles)

    # Initialise the weights
    weights = numpy.zeros(angles.size)

    # Compute the mean interval for normalisation
    if angles.size > 1:
        mean_interval = (angles[index[-1]] - angles[index[0]]) / (angles.size - 1)
    else:
        mean_interval = 1.0

    # Compute the angular weights
    for i in range(angles.size):
        sum_intervals = 0.0
        sum_weights = 0.0
        if i - 2 >= 0:
            sum_weights += 1.0 / 1.5
            sum_intervals += (1.0 / 1.5) * (angles[index[i - 1]] - angles[index[i - 2]])
        if i - 1 >= 0:
            sum_weights += 2.0
            sum_intervals += 2.0 * (angles[index[i]] - angles[index[i - 1]])
        if i + 1 < angles.size:
            sum_weights += 2.0
            sum_intervals += 2.0 * (angles[index[i + 1]] - angles[index[i]])
        if i + 2 < angles.size:
            sum_weights += 1.0 / 1.5
            sum_intervals += (1.0 / 1.5) * (angles[index[i + 2]] - angles[index[i + 1]])
        if sum_weights != 0:
            weights[index[i]] = sum_intervals / sum_weights
        else:
            weights[index[i]] = 1.0

    # Normalise the weights
    weights /= mean_interval

    # Return the weights
    return weights


def get_weights(angles, angular_weights=False):
    """
    Get the sinogram weights

    Params:
        angles (array): The array of angles
        angular_weights (bool): Use angular weighting

    Returns:
        array: List of sinogram weights

    """
    if angular_weights:
        weights = compute_angular_weights(angles)
    else:
        weights = numpy.ones(angles.shape)
    return weights


def reconstruct(
    tomogram,
    angles,
    reconstruction=None,
    centre=None,
    pixel_size=1,
    min_defocus=0,
    max_defocus=0,
    sinogram_order=False,
    transform=None,
    angular_weights=False,
    device="cpu",
    ncore=None,
    nchunk=None,
    gpu_list=None,
):
    """
    The sinogram is expected to be provided with dimensions as follows:

    - Stack of projections:
        ( THETA, Y, X ) - sinogram_order = False

    - Stack of sinograms:
        ( Y, THETA, X ) - sinogram_order = True

    - Stack of defocus corrected projections
        ( THETA, DEFOCUS, Y, X ) - sinogram_order = False

    - Stack of defocus corrected sinograms
        ( Y, DEFOCUS, THETA, X ) - sinogram_order = True

    """

    def initialise_sinogram(tomogram, sinogram_order):
        assert len(tomogram.shape) in [3, 4], "Dimensions must be 3 or 4"
        tomogram = tomogram.astype(dtype="float32", copy=False)
        if not sinogram_order:
            tomogram = numpy.swapaxes(tomogram, 0, -2)  # doesn't copy data
        return tomogram

    def initialise_reconstruction(reconstruction, sinogram):
        if reconstruction is None:
            shape = (sinogram.shape[0], sinogram.shape[-1], sinogram.shape[-1])
            reconstruction = numpy.zeros(shape, dtype="float32")
        else:
            assert (
                reconstruction.shape[0] == sinogram.shape[0]
            ), "Inconsistent reconstruction/sinogram shape"
            assert (
                reconstruction.shape[1] == sinogram.shape[-1]
            ), "Inconsistent reconstruction/sinogram shape"
            assert (
                reconstruction.shape[2] == sinogram.shape[-1]
            ), "Inconsistent reconstruction/sinogram shape"
        return reconstruction.astype(dtype="float32", copy=False)

    # Initialize sinogram
    sinogram = initialise_sinogram(tomogram, sinogram_order)

    # Initialise reconstruction
    reconstruction = initialise_reconstruction(reconstruction, sinogram)

    # Generate args for the algorithm.
    centre = get_centre(sinogram.shape, centre)

    # Get the weights
    weights = get_weights(angles, angular_weights)

    # Perform the reconstruction in multiple threads
    guanaco.detail.mp.reconstruction_dispatcher(
        sinogram,
        reconstruction,
        centre,
        angles,
        weights=weights,
        pixel_size=pixel_size,
        min_defocus=min_defocus,
        max_defocus=max_defocus,
        transform=transform,
        device=device,
        ncore=ncore,
        nchunk=nchunk,
        gpu_list=gpu_list,
    )

    # Return reconstruction
    return reconstruction


def reconstruct_file(
    input_filename,
    output_filename,
    corrected_filename="GUANACO_CORRECTED.dat",
    start_angle=None,
    step_angle=None,
    pixel_size=None,
    centre=None,
    energy=None,
    defocus=None,
    num_defocus=None,
    step_defocus=None,
    spherical_aberration=None,
    astigmatism=None,
    astigmatism_angle=None,
    phase_shift=None,
    device="cpu",
    ncore=None,
    transform=None,
    angular_weights=False,
    chunk_size=None,
    method="FBP_CTF",
    num_iter=None,
):
    """
    Do the reconstruction

    """
    start_time = time.time()

    def read_projection_metadata(infile):

        # Read the voxel size
        voxel_size = infile.voxel_size
        print("Voxel size: ", infile.voxel_size)

        # Read the angles
        if len(infile.data.shape) == 3:
            assert (
                infile.data.shape[0] == infile.extended_header.shape[0]
            ), "Inconsistent data/header shape"
            angles = numpy.zeros(infile.extended_header.shape[0], dtype=numpy.float32)
            for i in range(infile.extended_header.shape[0]):
                angles[i] = numpy.deg2rad(infile.extended_header[i]["Alpha tilt"])
                print("Image %d; angle %.4f deg" % (i + 1, numpy.rad2deg(angles[i])))
            min_defocus = 0
            max_defocus = 0
        else:
            infile.extended_header.shape = infile.data.shape[0:2]
            assert (
                infile.data.shape[0] == infile.extended_header.shape[0]
            ), "Inconsistent data/header shape"
            assert (
                infile.data.shape[1] == infile.extended_header.shape[1]
            ), "Inconsistent data/header shape"
            angles = numpy.zeros(infile.extended_header.shape[0], dtype=numpy.float32)
            for i in range(infile.extended_header.shape[0]):
                angles[i] = numpy.deg2rad(infile.extended_header[i, 0]["Alpha tilt"])
                print("Image %d; angle %.4f deg" % (i + 1, numpy.rad2deg(angles[i])))
            defocus = numpy.zeros(infile.extended_header.shape[1], dtype=numpy.float32)
            for i in range(infile.extended_header.shape[1]):
                defocus[i] = infile.extended_header[0, i]["Defocus"]
                print("Correction %d; defocus %.4f A" % (i + 1, defocus[i]))
            min_defocus = numpy.min(defocus)
            max_defocus = numpy.max(defocus)

        # Return metadata
        return angles, voxel_size, min_defocus, max_defocus

    def open_reconstruction_file(output_filename, shape, voxel_size):

        # Open the file
        outfile = mrcfile.new_mmap(
            output_filename, overwrite=True, mrc_mode=2, shape=shape
        )

        # Set the voxel size
        outfile.voxel_size = voxel_size

        # Return the handle
        return outfile

    # Open the input file
    print("Reading %s" % input_filename)
    with mrcfile.mmap(input_filename) as infile:

        # Get the projection data
        projections = infile.data

        # Get the projection metadata
        if start_angle is None or step_angle is None or pixel_size is None:
            angles, voxel_size, min_defocus, max_defocus = read_projection_metadata(
                infile
            )
            assert voxel_size["x"] == voxel_size["y"], "Pixel must be a square"
            pixel_size = voxel_size["x"]
        else:
            angles = numpy.deg2rad(
                start_angle + numpy.arange(projections.shape[0]) * step_angle
            )
            voxel_size = pixel_size
            min_defocus = 0
            max_defocus = 0

        # Get the pixel size
        print("Pixel size %d A" % pixel_size)

        # Set the reconstruction shape
        output_shape = (
            projections.shape[-2],
            projections.shape[-1],
            projections.shape[-1],
        )

        # Get the rotation centre
        centre = get_centre(projections.shape, centre, False)

        # Check if we have a corrected array
        if len(projections.shape) == 4:
            projections_are_corrected = True
            method = "FBP_CTF"
        else:
            projections_are_corrected = False

        # Get the corrected projections.
        if method in ["FBP_CTF"] and not projections_are_corrected:
            projections, min_defocus, max_defocus = guanaco.correct_projections(
                projections,
                centre,
                pixel_size,
                energy,
                defocus,
                num_defocus,
                step_defocus,
                spherical_aberration,
                astigmatism,
                astigmatism_angle,
                phase_shift,
                corrected_filename,
                device,
            )

        # Open the output file
        print("Writing reconstruction to %s" % output_filename)
        with open_reconstruction_file(
            output_filename, output_shape, voxel_size
        ) as outfile:

            # Get the reconstruction data
            reconstruction = outfile.data

            # Reconstruct
            if method in ["FBP_CTF"]:
                reconstruct(
                    projections,
                    angles,
                    reconstruction,
                    centre=centre,
                    pixel_size=pixel_size,
                    min_defocus=min_defocus,
                    max_defocus=max_defocus,
                    sinogram_order=False,
                    transform=transform,
                    angular_weights=angular_weights,
                    device=device,
                    ncore=ncore,
                )
            else:
                reconstruct_tomopy(
                    projections,
                    angles,
                    reconstruction,
                    centre=centre,
                    transform=transform,
                    device=device,
                    ncore=ncore,
                    method=method,
                    num_iter=num_iter,
                )

    print("Time: %.2f seconds" % (time.time() - start_time))


def reconstruct_tomopy(
    projections,
    angles,
    reconstruction,
    centre=None,
    transform=None,
    device="cpu",
    ncore=None,
    method="FBP",
    num_iter=200,
    output_rot90=False,
):
    """
    Do the reconstruction

    """
    try:
        import tomopy
    except ImportError:
        raise RuntimeError(
            "The reconstruction algorithm selected requires tomopy and astra"
        )

    # Set the device
    print("Using %s" % device)
    if device == "cpu":
        options = {"proj_type": "linear", "method": method}
    else:
        options = {"proj_type": "cuda", "method": "%s_CUDA" % method}
        if ncore == None:
            ncore = 1

    # Add number of iterations
    if method in ["SIRT", "SART", "CGLS"]:
        options["num_iter"] = num_iter

    # Transform the projections
    if transform == "minus":
        projections = -projections

    # Reconstruct
    reconstruction[:] = tomopy.recon(
        projections, angles, algorithm=tomopy.astra, options=options, ncore=ncore
    )
